import {Ping, PING_INTERVAL} from "./modules/ping.js";
import {Keyboard, SteeringWheel} from "./modules/input.js";
import {LogMemory} from "./modules/log.js"

/* eslint-env browser */
const queryString = window.location.search;
const urlParams = new URLSearchParams(queryString);
const DEBUG = (urlParams.get('debug') === 'true');  //ugly oh yes but wtf JS is any way fubared
const SAVEPING = (urlParams.get('saveping') === 'true')
const CARKEY = urlParams.get('carkey');
const ICEconfiguration = {
    iceServers: [{ urls: "stun:stun.l.google.com:19302" }],
};


// WEBSOCKET, connection between client and SSS.
let wsProto = "ws";
let wsHost = `${window.location.host}`
if (window.location.protocol == "file:" || wsHost.startsWith("localhost:")) {
    wsHost = "localhost:8765";
} else if (window.location.protocol != "http:") {
    wsProto += "s";
} else {
    wsHost = wsHost.replace(/:[0-9]+|$/,":8765")
}


const socket = new WebSocket(wsProto + `://${wsHost}/car/ws`);

// WEBRTC, connection between client and CAR.
const pc = new RTCPeerConnection(ICEconfiguration);

const LOG_MAX = 10
let logMessages = []

let log = (msg) => {
    console.log(msg);
    if (DEBUG){
        if (logMessages.length > LOG_MAX - 1) {logMessages.shift()}
        logMessages.push(msg + "<br>")
        document.getElementById("log").innerHTML = logMessages.join("");
    }
};


if (DEBUG) {
    // Maybe not needed but make the debug a bit easier (see real ip addresses)
    async function getMedia(constraints) {
        let stream = null;
        try { stream = await navigator.mediaDevices.getUserMedia(constraints); } catch (err) { log("Could not get media stream"); }
    }

    getMedia({ audio: true, video: true });
    document.getElementById("log").style.visibility = "visible";
}


// ------------------------------------------------------------------------------------------------------------
// WEBSOCKET, connection between client and SSS.
// ------------------------------------------------------------------------------------------------------------

// When connected to SSS
socket.onopen = () => {
    // FIXME: Check if CAR is there before sending offer...
    pc.createOffer().then((offer) => {
        pc.setLocalDescription(offer);
        socket.send(JSON.stringify(offer));
    });
};

// Messages recieved from SSS.
socket.onmessage = (e) => {
    let msg = JSON.parse(e.data);
    if (!msg) {
        return log("failed to parse msg");
    }

    if (msg.candidate) {
        // Got a candidate from SSS
        log("Remote:" + JSON.stringify(msg));
        pc.addIceCandidate(msg);
    } else {
        // Got a SDP from SSS
        pc.setRemoteDescription(msg);
    }
};

// ------------------------------------------------------------------------------------------------------------
// WEBRTC, connection between client and CAR.
// ------------------------------------------------------------------------------------------------------------

// Offer to receive 1 video tracks
pc.addTransceiver("video", { direction: "recvonly" });

// Called when a stream is found.
pc.ontrack = function (event) {
    var el = document.createElement(event.track.kind);
    el.srcObject = event.streams[0];
    el.muted = true;
    el.autoplay = true;
    el.controls = false;
    document.getElementById("Video").appendChild(el);
};

pc.onconnectionstatechange = (e) => log(pc.iceConnectionState);
pc.oniceconnectionstatechange = (e) => log(pc.iceConnectionState);
pc.onicegatheringstatechange = (e) => log(pc.iceGatheringState);

// Called when local candidates are found.
pc.onicecandidate = (e) => {
    if (e.candidate && e.candidate.candidate !== "") {
        log("Local:" + JSON.stringify(e.candidate));
        socket.send(JSON.stringify(e.candidate));
    }
};

// CAR controls.
let sendChannel = pc.createDataChannel("pilot");
sendChannel.binaryType = "arraybuffer";
sendChannel.onclose = () => {
    document.removeEventListener("keydown", keyboard.keyDown);
    document.removeEventListener("keyup", keyboard.keyUp);
}

let ping = null
let keyboard = null
let steeringWheel = null
let logmemory = null

sendChannel.onopen = () => {
    if (SAVEPING){
        logmemory = new LogMemory()
        document.getElementById("dlPingLogButton").style.display = 'block'
        document.getElementById("dlPingLogButton").onclick = logmemory.DownloadLog
    }
    
    ping = new Ping(sendChannel,CARKEY,logmemory)
    setInterval(() => ping.SendPing(),PING_INTERVAL) // Starts sending ping messages

    setRestartInfo(false)

    if (CarkeyExist()) {    //if we do not have the carkey do not setup the listeners
        keyboard = new Keyboard(CARKEY, sendChannel)
        document.addEventListener("keydown", (e) => keyboard.keyDown(e));
        document.addEventListener("keyup", (e) => keyboard.keyUp(e));
        window.addEventListener("gamepadconnected",(e) => {
            if (steeringWheel != null){return}
            steeringWheel = new SteeringWheel(CARKEY,sendChannel,e.gamepad)
        })
        window.addEventListener("gamepaddisconnected", (e) => {
            if (steeringWheel.GetIndex == e.gamepad.index){
            steeringWheel = null
            }
        })
    }
}

sendChannel.onmessage = (e) => {
    let dataView = new DataView(e.data);
    let decoder = new TextDecoder("utf8");
    let jsonmsg = JSON.parse(decoder.decode(dataView));
    let type = jsonmsg['Type'];
    let value = jsonmsg['Value'];
    log(decoder.decode(dataView));
    switch (type) {
        case "cario":
            let Throttle = value["Throttle"];
            let Steering = value["Steering"];
            let SteeringTrim = value["SteeringTrim"];
            let BreakMode = value["BreakMode"];
            let Active = value["Active"];
            if (Throttle > 0) {
                document.getElementById("Throttle-Gauge-forward").value = Throttle;
                document.getElementById("Throttle-Gauge-reverse").value = 0;
            } else {
                document.getElementById("Throttle-Gauge-forward").value = 0;
                document.getElementById("Throttle-Gauge-reverse").value = -1 * Throttle;
            };
            document.getElementById("wheel").style.transform = "rotate(" + Steering + "deg)";
            setRestartInfo(Active)

            UpdateKeyboard(BreakMode,SteeringTrim)
            UpdateSteeringWheel(BreakMode,SteeringTrim)
            if (SAVEPING){logmemory.SetThrottle(Throttle)}

            break;
        case "ping":
            ping.MeasureLatency(value)
            break;
        case "signalstrength":
            document.getElementById("Signalstrength").value = value;
            break;
        case "ping":
            ping.MeasureLatency(value)
            break;
        case "distance":
            UpdateDistance(value)
            break;
        default:
            log("Unknonw type: " + type);
    }
};


function CarkeyExist(){
    return typeof CARKEY === 'string' && !(CARKEY.trim().length === 0)
}

function setRestartInfo(active){
    if (CarkeyExist()){
        if (active) {document.getElementById("restart").style.display = 'none'}
        else {document.getElementById("restart").style.display = 'block'}
    }
}


function UpdateKeyboard(BreakMode, SteeringTrim){
    if (keyboard != null){
        keyboard.SetBreakMode(BreakMode)
        keyboard.SetSteeringTrim(SteeringTrim)
    }
}

function UpdateSteeringWheel(BreakMode, SteeringTrim){
    if (steeringWheel != null){
        steeringWheel.SetBreakMode(BreakMode)
        steeringWheel.SetSteeringTrim(SteeringTrim)
    }
}

function UpdateDistance(distance){
    let valueElem = document.getElementById("distance").children[0]
    if (distance >= 0) {valueElem.innerText = distance}
    else {valueElem.innerText = "-"}
    
    if (distance < 0) {valueElem.style.color = "black"}
    else if (distance <= 40) {valueElem.style.color = "darkred"}
    else if (distance <= 150) {valueElem.style.color = "darkorange"}
    else {valueElem.style.color = "darkgreen"}
}


