export {Keyboard, SteeringWheel}

class ControlDevice{
    Values = {
        forward: 0,
        break: 0,
        reverse: 0,
        left: 0,
        right: 0,
        throttle: 0,
        steering: 0,
        steeringTrim: 0,
        steeringTrimRight: 0,
        steeringTrimLeft: 0,
    }
    
    breakMode = false;


    constructor(carkey,sendChannel){
        this.carkey = carkey
        this.sendChannel = sendChannel
    }

    SendValue(cmd){
        switch (cmd){
            case "forward":           this.sendChannel.send(this.carkey + " forward " + String(this.Values["forward"])); break
            case "reverse":           this.sendChannel.send(this.carkey + " reverse " + String(this.Values["reverse"])); break
            case "break":             this.sendChannel.send(this.carkey + " break " + String(this.Values["break"])); break
            case "left":              this.sendChannel.send(this.carkey + " left " + String(this.Values["left"])); break
            case "right":             this.sendChannel.send(this.carkey + " right " + String(this.Values["right"])); break
            case "steeringTrimLeft":  this.sendChannel.send(this.carkey + " steeringTrimLeft " + String(this.Values["steeringTrimLeft"])); break
            case "steeringTrimRight": this.sendChannel.send(this.carkey + " steeringTrimRight " + String(this.Values["steeringTrimRight"])); break
            case "throttle":          this.sendChannel.send(this.carkey + " throttle " + String(this.Values["throttle"])); break
            case "steering":          this.sendChannel.send(this.carkey + " steering " + String(this.Values["steering"])); break
            case "steeringTrim":      this.sendChannel.send(this.carkey + " steeringTrim " + String(this.Values["steeringTrim"])); break
            case "reactivate":        this.sendChannel.send(this.carkey + " reactivate "); this.ResetThrottle(); break
            }


    }
    SetBreakMode(mode){
        this.breakMode = mode
    }
    
    SetSteeringTrim(value){
        this.Values["steeringTrim"] = value
    }
    
    ResetThrottle(){
            this.Values["forward"] = 0
            this.Values["reverse"] = 0
            this.Values["break"] = 0
            this.Values["throttle"] = 0
    }

}



class Keyboard extends ControlDevice {
    KeyMap = {
        ArrowUp: "forward",
        ArrowDown: "break",
        ArrowLeft: "left",
        ArrowRight: "right",
        KeyQ: "steeringTrimLeft",
        KeyE: "steeringTrimRight",
        Enter: "reactivate",
    }
    Controls = {
        "forward":           {step: 2,  repeatable: true, repeatInterval: 35,  resetUponRelease: true,  ctrl: "forward",      minVal: 0,   maxVal: 100},
        "reverse":           {step: 2,  repeatable: true, repeatInterval: 35,  resetUponRelease: true,  ctrl: "reverse",      minVal: 0,   maxVal: 100},
        "break":             {step: 8,  repeatable: true, repeatInterval: 35,  resetUponRelease: true,  ctrl: "break",        minVal: 0,   maxVal: 100},
        "left":              {step: 5,  repeatable: true, repeatInterval: 35,  resetUponRelease: true,  ctrl: "left",         minVal: 0,   maxVal: 100},
        "right":             {step: 5,  repeatable: true, repeatInterval: 35,  resetUponRelease: true,  ctrl: "right",        minVal: 0,   maxVal: 100},
        "steeringTrimLeft":  {step: -1, repeatable: true, repeatInterval: 500, resetUponRelease: false, ctrl: "steeringTrim", minVal: -10, maxVal: 10},
        "steeringTrimRight": {step: 1,  repeatable: true, repeatInterval: 500, resetUponRelease: false, ctrl: "steeringTrim", minVal: -10, maxVal: 10},
        "reactivate":	     {			repeatable: false, 											    ctrl: "reactivate"}
    }
    waitingKeys = []
    increasers = {}

    constructor(carkey, sendChannel){
        super(carkey,sendChannel)
        
    }

    increaseValue(cmd){
        if (this.Values[this.Controls[cmd].ctrl] + this.Controls[cmd].step > this.Controls[cmd].maxVal){
            this.Values[this.Controls[cmd].ctrl] = this.Controls[cmd].maxVal
        } else if (this.Values[this.Controls[cmd].ctrl] + this.Controls[cmd].step < this.Controls[cmd].minVal){
            this.Values[this.Controls[cmd].ctrl] = this.Controls[cmd].minVal
        } else {
            this.Values[this.Controls[cmd].ctrl] += this.Controls[cmd].step
        }
    }

    resetValue(cmd){
        this.Values[this.Controls[cmd].ctrl] = 0 
    }


    keyDown(event){
        // Read Keycode and corresponding control̈́
        let key = String(event.code)
        let ctrl = this.KeyMap[key]
        // Return if no corresponding control, or repeat. Prevent default function of the control
        if (event.repeat || !ctrl || this.increasers[ctrl] != null){return;}
        event.preventDefault();
  
        // If control is blocked, place it in waiting list. Else, output 
        if (this.ctrlBlocked(ctrl) == false){
            this.startCtrl(ctrl)
        } else {
            this.waitingKeys.push(key)
        }
    }

 
    startCtrl(ctrl){
		if (this.Controls[ctrl].step){this.increaseValue(ctrl)}
        this.SendValue(this.Controls[ctrl].ctrl)

        if (this.Controls[ctrl].repeatable){
            let interval = this.Controls[ctrl].repeatInterval
            let increaser = setInterval(() => {
				if (this.Controls[ctrl].step){this.increaseValue(ctrl)}
                this.SendValue(this.Controls[ctrl].ctrl)
            },interval)
            this.increasers[ctrl] = increaser 
        } 
    }


    keyUp(event){
        // Read Keycode and corresponding control
        let key = String(event.code)
        let ctrl = this.KeyMap[key]

        // Return if no corresponding control, or repeat. Prevent default function of the control
        if (event.repeat || !ctrl){return;}
        event.preventDefault();
        

        // Remove key from waiting list if waiting
        if (this.waitingKeys.includes(key)){
            this.waitingKeys.splice(this.waitingKeys.indexOf(key))
            return
        }
        
        // Stop repeating key, if repeatable
        if (this.Controls[ctrl].repeatable){
            clearInterval(this.increasers[ctrl])
            this.increasers[ctrl] = null
            if (this.Controls[ctrl].resetUponRelease){
                this.resetValue(ctrl)
                this.SendValue(ctrl)
            }
        }

        // Start output if waiting key blocked by released key
        this.waitingKeys.forEach(key => {
            ctrl = this.KeyMap[key]

            if (this.ctrlBlocked(ctrl) == false){
                this.startCtrl(ctrl)
                this.waitingKeys.splice(this.waitingKeys.indexOf(key))
            }
        })
    }

    ctrlBlocked(ctrl){
        return (
           ctrl == "forward"   && this.increasers["reverse"]   != null
        || ctrl == "forward"   && this.increasers["break"]     != null
        || ctrl == "reverse"   && this.increasers["forward"]   != null
        || ctrl == "break"     && this.increasers["forward"]   != null
        || ctrl == "left"      && this.increasers["right"]     != null
        || ctrl == "right"     && this.increasers["left"]      != null
        || ctrl == "trimLeft"  && this.increasers["trimRight"] != null
        || ctrl == "trimRight" && this.increasers["trimLeft"]  != null) 
    }

    SetBreakMode(mode){
        this.breakMode = mode
        if (mode == true) {this.KeyMap.ArrowDown = "break"}
        else {this.KeyMap.ArrowDown = "reverse"}
    }
    
}

class Gamepad extends ControlDevice {
    index;
    STANDARD_DRIFTRANGE = 0.05;
    buttonMap = {}
    buttonVals = {}
    buttonState = {}
    axesMap = {}
    axesVal = {}
    gpTimer = null
    gpInterval = 35 

    constructor(carkey, sendChannel, gamepad){
        super(carkey,sendChannel)
        this.index = gamepad.index
        
        this.gpTimer = setInterval(() => {
            let gamepad = navigator.getGamepads()[this.index]
            this.Update(gamepad)
        },this.gpInterval)
        

    }

    Inactivate(){
        clearInterval(this.gpTimer)
    }    

    GetIndex(){
        return this.index
    }


    ReadAxis(gamepad, axis, invert = false, driftRange = this.STANDARD_DRIFTRANGE, idle = 0, fromScale = [-1,1], toScale = [-100,100]){
        // (1) Read the axis
        var value = gamepad.axes[axis]

        // (2) Remove jitter
        value = this.removeJitter(value,driftRange,idle)     
        
        // (3) Limit the number
        value = this.limitRange(value,fromScale[0],fromScale[1])
 
        // (4) Invert (if wanted)
        if (invert){value = -value}

        // (5) Rescale the number
        value = this.rescaleNumber(value,fromScale[0], fromScale[1], toScale[0], toScale[1])
        
        // (6) Limit number of decimals
        value = Math.trunc(value)

        return value
    }

    ReadButton(gamepad, button){
        return gamepad.buttons[this.buttonMap[button]].pressed
        
    }

    ReadAnalogButton(gamepad, button){
        return gamepad.buttons[this.buttonMap[button].value]
    }



    buttonDown(gamepad,buttonName){
        return !this.buttonState[buttonName] && gamepad.buttons[this.buttonMap[buttonName]].pressed
    }    

    buttonUp(gamepad,buttonName){
        return this.buttonState[buttonName] && !gamepad.buttons[this.buttonMap[buttonName]].pressed
    }    

    SaveGamepadState(gamepad){
        Object.keys(this.buttonMap).forEach(button => {
            this.buttonState[button] = gamepad.buttons[this.buttonMap[button]].pressed
            this.buttonVals[button] = gamepad.buttons[this.buttonMap[button]].value
        })
        Object.keys(this.axesMap).forEach(axis => {
            this.axesVal[axis] = gamepad.axes[this.axesMap[axis]].value
        })
    }

    rescaleNumber(input, fromMin, fromMax, toMin, toMax){
        return (input - fromMin) * (toMax - toMin) / (fromMax - fromMin) + toMin
    }

    limitRange(input, min,max){
        if (input < min) {return min}
        else if (input > max) {return max}
        else {return input}
    }

    removeJitter(input, range, mid = 0){
        if (input > mid - range && input < mid + range){
            return mid
        } else {
            return input
        }
    }
}

class SteeringWheel extends Gamepad {
    buttonMap = { 
        SteeringTrimLeft: 11,
        SteeringTrimRight: 10,
        Reactivate: 3,
    }

    // TODO: Maybe generate this from button map
    buttonState = {
        SteeringTrimLeft: false,
        SteeringTrimRight: false,
        Reactivate: false,
    }

    axesMap = { 
        Throttle: 2,
        Break: 5,
        Reverse: 1,
        Steering: 0,
    }


    constructor(carkey, sendChannel, gamepad){
        super(carkey,sendChannel,gamepad)

    }

    Update(gamepad){

        if (this.Values.break == 0 && this.Values.reverse == 0){
            let newThrottle = this.ReadAxis(gamepad, this.axesMap.Throttle,true,undefined,-1,undefined,[0,100])
            if (newThrottle != this.Values.forward) {this.Values.forward = newThrottle; this.SendValue("forward");} 
        }  

        if (this.Values.forward == 0 && this.Values.reverse == 0) {
            let newBreak = this.ReadAxis(gamepad, this.axesMap.Break,true,undefined,-1,undefined,[0,100])
            if (newBreak != this.Values.break) {this.Values.break = newBreak; this.SendValue("break");} 
        }
        
        if (this.Values.forward == 0 && this.Values.break == 0){
            let newReverse = this.ReadAxis(gamepad, this.axesMap.Reverse,true,undefined,-1,undefined,[0,100])
            if (newReverse != this.Values.reverse) {this.Values.reverse = newReverse; this.SendValue("reverse");}
        }
        
        let newSteering = this.ReadAxis(gamepad, this.axesMap.Steering,false,undefined,0,[-0.25,0.25],undefined)
        if (newSteering != this.Values.steering) {this.Values.steering = newSteering; this.SendValue("steering");}


        
        if (this.buttonDown(gamepad, 'SteeringTrimLeft') && !this.buttonDown(gamepad, 'SteeringTrimRight')){
            let newSteeringTrim = this.Values.steeringTrim - 1
            newSteeringTrim = this.limitRange(newSteeringTrim,-10,10)
            if (newSteeringTrim != this.Values.steeringTrim) {this.Values.steeringTrim = newSteeringTrim; this.SendValue("steeringTrim");}
        }

        if (this.buttonDown(gamepad, 'SteeringTrimRight') && !this.buttonDown(gamepad, 'SteeringTrimLeft')){
            let newSteeringTrim = this.Values.steeringTrim + 1
            newSteeringTrim = this.limitRange(newSteeringTrim,-10,10)
            if (newSteeringTrim != this.Values.steeringTrim) {this.Values.steeringTrim = newSteeringTrim; this.SendValue("steeringTrim");}
        }
        
        if (this.buttonDown(gamepad, 'Reactivate')){
            this.SendValue("reactivate")
        }
        
        this.SaveGamepadState(gamepad)
    }

}
